/* MicroLens.cpp   =======================================================*

*Important
===================================================
Lenslet to Multiview Converter Version 1.0

Copyright: ONLY Available for ACADEMIC and MPEG Standardization Purposes
Production: Nagoya University


*Description
===================================================
Given a set of Raw video captured by multi-focused plenoptic camera (i.e. Ratrix camera, https://raytrix.de/) 
and its microlens calibration data, the provided source code can render multi-view video.

This tool is implemented by:

Fujii Laboratory
Department of Information and Communication Engineering
Graduate School of Engineering
Nagoya University


*Term of Use
===================================================
Term of Use: ANY kind of academic publication, and report using this tool MUST refer to the following TWO references.

[1] Shu Fujita, Sho Mikawa, Mehrdad Teratani(Panahpourtehrani), Keita Takahashi, and Toshiaki Fujii 
“Extracting Multi-View Images from Multi-Focused Plenoptic Camera,” 
International Workshop on Advanced Image Technology (IWAIT)/International Forum on Medical Imaging in Asia (IFMIA), 
in SPIE Digital Library, Singapore, January 2019

[2] Mehrdad Teratani(Panahpourtehrani), Shu Fujita, Wenzhe Ouyang, Keita Takahashi, Toshiaki Fujii, 
"3D Imaging System Using Multi-Focus Plenoptic Camera and Tensor Display," 
2018 International Conference on 3D Immersion (IC3D2018) in IEEE Xplore Digital Library, Brussels, Belgium, December 2018 

===================================================
* Input: Raw image (.png), Calibration (.xml)
*
* @file		RenderingRxMV > main.cpp
* @author	S. Mikawa 
* @date		2018/08/31
*======================================================================*/

#include "MicroLens.h"

MicroLens::MicroLens()
{
}

MicroLens::~MicroLens()
{
}

void MicroLens::initFromRay()
{

}

void MicroLens::initFromXML(Mat rawImage, string filename)
{
	//Read Raytrix XML
	ptree tree;
	read_xml(filename, tree);

	diameter = tree.get<double>("RayCalibData.diameter");
	rotation = tree.get<double>("RayCalibData.rotation");
	radius = diameter * 0.5;
	center.x = rawImage.cols * 0.5 + tree.get<double>("RayCalibData.offset.x");
	center.y = rawImage.rows * 0.5 - tree.get<double>("RayCalibData.offset.y");

	//Set Lens Type Offset
	BOOST_FOREACH(const ptree::value_type& child, tree.get_child("RayCalibData"))
		if (child.first == "lens_type")
		{
			offset[child.second.get<int>("<xmlattr>.id")].x = child.second.get<float>("offset.x");
			offset[child.second.get<int>("<xmlattr>.id")].y = child.second.get<float>("offset.y");
		}

	//Calculation Number of Micro Lens
	int upNum = (int)((center.y - radius) / (radius * sqrt(3)));
	int downNum = (int)((rawImage.rows - (center.y + radius)) / (radius * sqrt(3)));
	num.y = upNum + 1 + downNum;
	double leftNum = (center.x - radius) / diameter;
	double rightNum = (rawImage.cols - (center.x + radius)) / diameter;

	if (((leftNum - (int)leftNum) > 0.5) || ((rightNum - (int)rightNum) >= 0.5))
		num.x = (int)leftNum + 1 + (int)rightNum;
	else
		num.x = (int)leftNum + (int)rightNum;
	num.x = (int)(num.x / 3);

	//Calculation subGridRefPos
	for (int type = 0; type < 3; type++)
	{
		double refPos = center.x + diameter * -(offset[type].x + offset[type].y);	//On the assumption that either offset[].x or offset[].y is Zero.
		double dis = diameter * 3;
		subGridRefPos[upNum % 2][type].x = refPos - (int)((refPos - radius) / dis) * dis;
		refPos = center.y;
		dis = diameter * sqrt(3);
		subGridRefPos[upNum % 2][type].y = refPos - (int)((refPos - radius) / dis) * dis;

		refPos = center.x + diameter * -(offset[type].x + offset[type].y) - (radius + diameter);
		dis = diameter * 3;
		subGridRefPos[abs(upNum % 2 - 1)][type].x = refPos - (int)((refPos - radius) / dis) * dis;
		refPos = center.y - radius * sqrt(3);
		dis = diameter * sqrt(3);
		subGridRefPos[abs(upNum % 2 - 1)][type].y = refPos - (int)((refPos - radius) / dis) * dis;
	}
}

void MicroLens::showParam()
{
	cout << "==========MicroLens Parameter==========" << endl;
	cout << "diameter: " << diameter << endl;
	cout << "rotation: " << rotation << endl;
	cout << "num: " << num << endl;
	cout << "center: " << center << endl;
	for (int type = 0; type < 3; type++)
		cout << "offset[" << type << "]: " << offset[type] << endl;
	for (int gridNum = 0; gridNum < 2; gridNum++)
		for (int type = 0; type < 3; type++)
			cout << "subGridRefPos[" << gridNum << "][" << type << "]: " << subGridRefPos[gridNum][type] << endl;
	cout << "=======================================" << endl;
}

void MicroLens::draw(Mat& rawImage, lensType type, int patchSize)
{
	Mat drawImage;// = rawImage.clone();
	cvtColor(rawImage, drawImage, CV_GRAY2RGB);
	Scalar color[5] = { Scalar(0, 0, 255), Scalar(0, 255, 0), Scalar(255, 0, 0), Scalar::all(255), Scalar::all(0) };

	//Image Center Line
	line(drawImage, Point2d(drawImage.cols * 0.5, 0), Point2d(drawImage.cols * 0.5, drawImage.rows), color[BLUE], 2, 8, 0);
	line(drawImage, Point2d(0, drawImage.rows * 0.5), Point2d(drawImage.cols, drawImage.rows * 0.5), color[BLUE], 2, 8, 0);

	//Each Lens Type
	int shift = 4;
	Point2d lensCenter;
	lensType firstType = type;
	lensType endType = type;
	if (type == TYPE_ALL)
	{
		firstType = TYPE1;
		endType = TYPE3;
	}
	for (int typeCnt = firstType; typeCnt <= endType; typeCnt++)
	{
		for (int y = 0; y < num.y; y++)
		{
			for (int x = 0; x < num.x; x++)
			{
				lensCenter.x = subGridRefPos[y % 2][typeCnt].x + diameter * 3 * x;
				lensCenter.y = subGridRefPos[0][0].y + radius * sqrt(3) * y;

				if (patchSize != 0)
				{
					double patchRatio = 23.0 / patchSize;
					lensCenter *= patchRatio;
				}

				circle(drawImage, (1 << shift) * lensCenter, (1 << shift) * (int)radius, color[typeCnt], 1, 8, shift);
			}
		}
	}

	//Lens Frame Center Point	
	circle(drawImage, (1 << shift) * center, (1 << shift) * (int)radius, Scalar(0, 0, 255), 1, 8, shift);

	//Show Image
	cv::imshow("drawImage", drawImage);
	imwrite("drawImage.png", drawImage);
}

lensType MicroLens::getRefLensType(int subGridNum, lensPos pos)
{
	double temp = min(subGridRefPos[subGridNum][TYPE1].x, min(subGridRefPos[subGridNum][TYPE2].x, subGridRefPos[subGridNum][TYPE3].x));

	if (pos == MIDDLE)
		temp += diameter;
	if (pos == RIGHT)
		temp += diameter * 2;

	for (int type = TYPE1; type < TYPE_ALL; type++)
		if ((int)temp == (int)subGridRefPos[subGridNum][type].x)
			return (lensType)type;
	return TYPE1;
}