/* MultiViewImage.h ======================================================================*

*Important
===================================================
Lenslet to Multiview Converter Version 1.0

Copyright: ONLY Available for ACADEMIC and MPEG Standardization Purposes
Production: Nagoya University


*Description
===================================================
Given a set of Raw video captured by multi-focused plenoptic camera (i.e. Ratrix camera, https://raytrix.de/) 
and its microlens calibration data, the provided source code can render multi-view video.

This tool is implemented by:

Fujii Laboratory
Department of Information and Communication Engineering
Graduate School of Engineering
Nagoya University


*Term of Use
===================================================
Term of Use: ANY kind of academic publication, and report using this tool MUST refer to the following TWO references.

[1] Shu Fujita, Sho Mikawa, Mehrdad Teratani(Panahpourtehrani), Keita Takahashi, and Toshiaki Fujii 
“Extracting Multi-View Images from Multi-Focused Plenoptic Camera,” 
International Workshop on Advanced Image Technology (IWAIT)/International Forum on Medical Imaging in Asia (IFMIA), 
in SPIE Digital Library, Singapore, January 2019

[2] Mehrdad Teratani(Panahpourtehrani), Shu Fujita, Wenzhe Ouyang, Keita Takahashi, Toshiaki Fujii, 
"3D Imaging System Using Multi-Focus Plenoptic Camera and Tensor Display," 
2018 International Conference on 3D Immersion (IC3D2018) in IEEE Xplore Digital Library, Brussels, Belgium, December 2018 

===================================================
* Class for generating raytrix multi-view image
*
* @file		RenderingRxMV > MultiViewImage.h
* @author	S. Mikawa and S. Fujita
* @date		2018/10/dd
*======================================================================*/

#pragma once
#include <opencv2/opencv.hpp>
#include <vector>
#include <stdlib.h>
#include "MicroLens.h"
#include "StereoMatching.h"

#define DIVISION_NUM 6	//For DEPTH_BASE_V3

using namespace cv;
using namespace std;

enum patchEstimationMode { LAPLACIAN_BASE_ALL, LAPLACIAN_BASE_CENTER, DEPTH_BASE_V1, DEPTH_BASE_V2};
enum mergeMode { AVERAGE, GAUSSIAN, BLEND};
enum lensAlignment {NMF, FMN, NFM, FNM};

class MultiViewImage
{
private:
	//Variable for Debug
	int trackbarLensType;
	int trackbarPatchSize;
	int trackbarPatchEstimationMode;
	int trackbarValue;
	Point trackbarMLNum;
	Point trackbarViewNum;

	//Variable
	MicroLens ml;
	RaytrixStereoMatching* rsm;
	int imageNum;
	int subImageOverlap;
	int patchInterval;
	int lensTypeOffset;
	Point centerView;
	Mat rawImage;
	Mat maskImage;
	Mat maskLapImage;
	Mat dividedMaskImage[DIVISION_NUM];
	Mat patchSize[TYPE_ALL];
	Mat dividedPatchSize[TYPE_ALL][DIVISION_NUM];
	Mat patchMap[TYPE_ALL];
	Mat patchVisualizationMap[TYPE_ALL];
	Size multiViewImageSize;
	vector<int> patchSizeList;
	vector<vector<Mat>> resizeRawImageList;
	vector<vector<Mat>> multiViewImage[TYPE_ALL];

	Mat edgeStd[TYPE_ALL];
	Mat edgeStdMap[TYPE_ALL];
	Mat edgeStdVisualizationMap[TYPE_ALL];

	//Method
	int getPatchListIndex(int patchSize);
	void generateMask(Mat& maskImage);
	void generateResizeRawImage();
	bool checkRoiRange(Mat& srcImage, Rect roi);
	void alphaBlend(Mat& srcImage, Mat& destImage, Mat& maskImage, double alpha);
	void getRoiImage(Mat& srcImage, Mat& destImage, Mat& maskImage, Rect roi);
	void setRoiImage(Mat& srcImage, Mat& destImage, Mat& maskImage, Rect roi);
	void patchVisualization(lensType type = TYPE_ALL);
	void generatePatchMap(int mode);
	void getShiftImage(Mat& srcImage, Mat& dstImage, Point shift);
	void rendering(bool isSpecificPatchSize, bool isSpecificView, int patchSize = 0, Point view = Point(0, 0));
	void shiftMultiViewImage();
	void merge(vector<vector<Mat>>& destImage, int mmode, int lmode);
	void correctionImage(vector<vector<Mat>>& destImage, bool isTrimming = true, bool isRotate = true);
	void correctionImage(Mat& destImage, bool isTrimming = true, bool isRotate = true);
	void visualMode();

	//Patch Estimation Mode
	void laplacianBaseEstimation(bool isSpecificView = false, Point view = Point(0, 0));
	void depthBaseEstimationV1();
	void depthBaseEstimationV2();
	void depthBaseEstimationV3();


public:
	//Constructor and Destructor
	MultiViewImage(MicroLens ml, const int imageNum, int microlensOverlap);
	~MultiViewImage();

	//Method
	void saveMultiViewImage_RGB(char* folderName, vector<vector<Mat>>& srcImage);
	void saveMultiViewImage_YUV(char* folderName, vector<vector<Mat>>& srcImage);
	void saveEachLensTypeCenterView(char* folderName);
	void savePatchMap(char* folderName);

	//Core
	void operator()(Mat srcImage, vector<vector<Mat>>& destImage, int pmode = LAPLACIAN_BASE_ALL, int mmode = AVERAGE, int lmode = NMF);
};

