/* main.cpp ==============================================================*

* Important
===================================================
Lenslet to Multiview Converter Version 1.0

Copyright: ONLY Available for ACADEMIC and MPEG Standardization Purposes
Production: Nagoya University


* Description
===================================================
Given a set of Raw video captured by multi-focused plenoptic camera (i.e. Ratrix camera,https://raytrix.de/) 
and its microlens calibration data, the provided source code can render multi-view video.

This tool is implemented by:

Fujii Laboratory
Department of Information and Communication Engineering
Graduate School of Engineering
Nagoya University


* Term of Use
===================================================
Term of Use: ANY kind of academic publication, and report using this tool MUST refer to the following TWO references.

[1] Shu Fujita, Sho Mikawa, Mehrdad Teratani(Panahpourtehrani), Keita Takahashi, and Toshiaki Fujii 
“Extracting Multi-View Images from Multi-Focused Plenoptic Camera,” 
International Workshop on Advanced Image Technology (IWAIT)/International Forum on Medical Imaging in Asia (IFMIA), 
in SPIE Digital Library, Singapore, January 2019

[2] Mehrdad Teratani(Panahpourtehrani), Shu Fujita, Wenzhe Ouyang, Keita Takahashi, Toshiaki Fujii, 
"3D Imaging System Using Multi-Focus Plenoptic Camera and Tensor Display," 
2018 International Conference on 3D Immersion (IC3D2018) in IEEE Xplore Digital Library, Brussels, Belgium, December 2018 

===================================================
* Input: Raw image (.png), Calibration (.xml)
*
* @file		RenderingRxMV > main.cpp
* @author	S. Mikawa, S. Fujita, and W. Ouyang
* @date		2018/08/31
*======================================================================*/


//Head file and Standard library
#include "MultiViewImage.h"
#include "MicroLens.h"
#include <unistd.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>  
#include <sys/types.h>

//openc library, you can use "#include <opencv2/opencv.hpp>" instead. 
#define CV_LIB_PREFIX comment(lib, "opencv_"

#define CV_LIB_VERSION CVAUX_STR(CV_MAJOR_VERSION)\
    CVAUX_STR(CV_MINOR_VERSION)\
    CVAUX_STR(CV_SUBMINOR_VERSION)

#ifdef _DEBUG
#define CV_LIB_SUFFIX CV_LIB_VERSION "d.lib"
#else
#define CV_LIB_SUFFIX CV_LIB_VERSION ".lib"
#endif

#define CV_LIBRARY(lib_name) CV_LIB_PREFIX CVAUX_STR(lib_name) CV_LIB_SUFFIX

#pragma CV_LIBRARY(core)
#pragma CV_LIBRARY(highgui)
#pragma CV_LIBRARY(imgcodecs)
#pragma CV_LIBRARY(imgproc)

int main(int argc, char* argv[])
{
	char Calibration_xml[128];
	char RawImage_Path[128];
	char RawImage_fullPath[128];
	char Output_Path[128];
	char Output_fullPath[128];
	int viewNum, microlensOverlap, pmode, mmode, lmode, input_model, output_model, start_frame, end_frame, height, width;

	//Make sure that you specify parameter file (param.cfg). 
	if (argc != 2)
	{
		cout<<"command line error"<<endl;
		exit(0);
	}

	//Scan Parameter file
	FILE *fp;
	char s[256];

	fp = fopen(argv[1], "r");

	fscanf(fp, "%s\t%d", s, &viewNum);
	fscanf(fp, "%s\t%d", s, &microlensOverlap);
	fscanf(fp, "%s\t%d", s, &pmode);
	fscanf(fp, "%s\t%d", s, &mmode);
	fscanf(fp, "%s\t%d", s, &lmode);
	fscanf(fp, "%s\t%s", s, Calibration_xml);
	fscanf(fp, "%s\t%s", s, RawImage_Path);
	fscanf(fp, "%s\t%s", s, Output_Path);
	fscanf(fp, "%s\t%d", s, &input_model);
	fscanf(fp, "%s\t%d", s, &output_model);
	fscanf(fp, "%s\t%d", s, &start_frame);
	fscanf(fp, "%s\t%d", s, &end_frame);
	fscanf(fp, "%s\t%d", s, &height);
	fscanf(fp, "%s\t%d", s, &width);

	fclose(fp);

	cout<<Calibration_xml<<endl;
	cout<<viewNum<<' '<<pmode<<' '<<mmode<<' '<<lmode<<endl;

	//Timing function
	time_t start,stop;
	start = time(NULL);
	Mat rawImage;

	//Processing
	for(int i=start_frame; i<=end_frame; i++){
		sprintf(RawImage_fullPath, RawImage_Path, i);
		//sprintf(RawImage_fullPath, "%s_Processed.png", RawImage_fullPath);
		cout<<RawImage_fullPath<<endl;
		if (input_model==0)
		{
		    //cvtColor(rawImage, rawImage, COLOR_YUV2BGR_I420);
			rawImage = imread(RawImage_fullPath, 1);
		}
		else{

			//Read YUV file including 1 frame data.
			FILE *fileIn = fopen(RawImage_fullPath,"rb+");
			
			//Specify height and width in Parameter file.
			const int framesize = height * width * 3/2;
			unsigned char* pYuvBuf = new unsigned char[framesize];

			fread(pYuvBuf, framesize*sizeof(unsigned char), 1, fileIn);
			Mat yuvImg;
			yuvImg.create(height*3/2, width, CV_8UC1);
			memcpy(yuvImg.data, pYuvBuf, framesize*sizeof(unsigned char));

			cvtColor(yuvImg, rawImage, COLOR_YUV2BGR_I420);
		}

		//Set Micro Lens Parameter
		MicroLens ml;
		ml.initFromXML(rawImage, Calibration_xml);
		MultiViewImage multiViewApp(ml, viewNum, microlensOverlap);

		vector<vector<Mat>> multiViewImage;

		//Get Multi-view Image
		multiViewApp(rawImage, multiViewImage, pmode, mmode, lmode);

		//Save Image		
		if (output_model==1){
			cout<<"Saving Images with YUV format"<<endl;

			sprintf(Output_fullPath, Output_Path, i);
			cout<<Output_fullPath<<endl;

			mkdir(Output_fullPath, 0777);
			chmod(Output_fullPath, 0777);
			
			multiViewApp.saveMultiViewImage_YUV(Output_fullPath, multiViewImage);
		}
		else{		
			cout<<"Saving Images with RGB format"<<endl;

			sprintf(Output_fullPath, Output_Path, i);
			cout<<Output_fullPath<<endl;
			
			mkdir(Output_fullPath, 0777);
			chmod(Output_fullPath, 0777);

			multiViewApp.saveMultiViewImage_RGB(Output_fullPath, multiViewImage);
		}

		//Save LensType Images and Patch map.
		multiViewApp.saveEachLensTypeCenterView(Output_fullPath);
		multiViewApp.savePatchMap(Output_fullPath);
	}

	stop = time(NULL);
	printf("Use Time:%ld\n",(stop-start));
	return 0;
}